import sys
import time
import json
import struct
import random
import microstacknetwork
import microstacknetwork.packets
import microstacknetwork.handshake
import microstacknetwork.radios.nrf24l01
import microstacknode.accelerometer.mma8452q
import microstacknode.gps.l80gps

import logging
logging.basicConfig(level=logging.DEBUG)


ACTION_REQ_INTERVAL = 0.1
BEACON_ADDRESS = 0x1234567890
# NODE_INIT_ADDRESS = int(0xffffffffff * random.random())
NODE_INIT_ADDRESS = 0x3cc23f3d5e


channel_values = [0xAB,
                  0xCD,
                  ]


class NodeFailedHandshakeError(Exception):
    pass


class MicrostackNodeChannel(object):
    """A channel on a Microstack Node."""

    def __str__(self):
        return self.name

    def get(self):
        return None

    def set(self, value):
        pass

    def get_block(self, length):
        """Some sensors are spread accross several channels becasue of the
        1 Byte limit. The channels can be read with `get_block` but it is
        sometimes inefficient to read each channel seperately and then combine
        then later. This is becasue the large sensor will be split up for each
        read. To illustrate:

            Get Block 4 <- Read channel 0 <- Read whole sensor Sensor (4B)
                           Read channel 1 <- Read whole sensor Sensor (4B)
                           Read channel 2 <- Read whole sensor Sensor (4B)
                           Read channel 3 <- Read whole sensor Sensor (4B)
            = 4 sensor reads when we only require 1

        Use this funciton to define a default channel read_block action.
        If this is not defined then the default behaviour of reading each
        channel individually is used:

            Get Block 4 <- Read channels 0-3 <- Read whole sensor Sensor (4B)
        """
        raise NotImplementedError()

    def set_block(self, values):
        raise NotImplementedError()


class AccelerometerX(MicrostackNodeChannel):

    name = "accel_x"

    def __init__(self, accelerometer):
        self.accelerometer = accelerometer

    def get(self):
        # get 8-bit raw values
        x, y, z = self.accelerometer.get_xyz(raw=True, res12=False)
        return x

    def get_block(self, length):
        xyz = self.accelerometer.get_xyz(raw=True, res12=False)
        return xyz[:length]


class AccelerometerY(MicrostackNodeChannel):

    name = "accel_y"

    def __init__(self, accelerometer):
        self.accelerometer = accelerometer

    def get(self):
        # get 8-bit raw values
        x, y, z = self.accelerometer.get_xyz(raw=True, res12=False)
        return y


class AccelerometerZ(MicrostackNodeChannel):

    name = "accel_z"

    def __init__(self, accelerometer):
        self.accelerometer = accelerometer

    def get(self):
        # get 8-bit raw values
        x, y, z = self.accelerometer.get_xyz(raw=True, res12=False)
        return z


class GPSBaseChannel(MicrostackNodeChannel):
    """Base class for latitude and longitude (they're basically the same).

    Latitude/Longitude is represented in four parts:

        N/S or E/W direction | degrees | minutes | seconds
        N                      31        10        17

    Latitude goes from 90S (-90) to 90N (+180) degrees. Longitude goes
    from 180W (-180) to 180E (+180) degrees Each degree has 60 minutes,
    each minute has 60 seconds.

    Latitude reading is ddmm.mmmm,D (d=degree, m=minutes, D=direction):

        3110.2908,N
        to: 31 degrees 10.2908 minutes North
        to: 31.17151 North
        to: +31.17151 (-ve if South)

    Longitude reading is dddmm.mmmm,D (d=degree, m=minutes, D=direction):

        12123.2348,E
        to: 121 degrees 23.2348 minutes East
        to: 121.38724 East
        to: +121.38724 (-ve if West)

    GPSLatitude/GPSLongitude returns a signed floating point latitude
    value in degrees. There are four bytes to the floating point value
    and so this is stored in four channels. Specfiy which channel with
    `index`.

    Might need to use JSPack to unpack these values in Javascript.

        https://code.google.com/p/jspack/

    """

    prop_name = "ERROR"

    def __init__(self, gps, index):
        self.gps = gps
        self.index = index
        self.name = "gps_{}_{}".format(self.prop_name, self.index)

    def get(self):
        return self.get_bytes()[self.index]

    def get_block(self, length):
        return [b for b in self.get_bytes()[:length]]

    def get_bytes(self):
        gpgll, checksum = microstacknode.gps.l80gps.gpgll_as_dict(
            self.gps.gpgll)
        degrees = degrees_and_minutes_to_degrees(gpgll[self.prop_name])
        if self.neg_dir in gpgll[self.dir_prop_name]:
            degrees *= -1
        return struct.pack("f", degrees)  # four byte, signed float


class GPSLatitude(GPSBaseChannel):
    prop_name = "latitude"
    dir_prop_name = "ns"
    neg_dir = "S"


class GPSLongitude(GPSBaseChannel):
    prop_name = "longitude"
    dir_prop_name = "ew"
    neg_dir = "W"


class DummyChannel(MicrostackNodeChannel):

    def __init__(self, name="dummy"):
        self.value = 0
        self.name = name

    def get(self):
        return self.value

    def set(self, v):
        self.value = v


class MicrostackNode(object):

    def __init__(self, node_address=None, server_address=None):
        self.radio = microstacknetwork.radios.nrf24l01.NRF24L01()
        self.radio.init()
        self.node_address = node_address
        self.server_address = server_address

        accelerometer = microstacknode.accelerometer.mma8452q.MMA8452Q()
        accelerometer.init()
        self.gps = microstacknode.gps.l80gps.L80GPS()
        # self.gps = microstacknode.gps.l80gps.L80GPS(open('/dev/ttyAMA0'))
        self.gps.start()  # start reading the GPS stream
        self.channels = [AccelerometerX(accelerometer),
                         AccelerometerY(accelerometer),
                         AccelerometerZ(accelerometer)]
        self.channels.extend([GPSLatitude(self.gps, i) for i in range(4)])
        self.channels.extend([GPSLongitude(self.gps, i) for i in range(4)])
        self.channels.extend([DummyChannel("dummy{}".format(i))
                              for i in range(3)])

        # for the service_action method. We only need to initialise dict once.
        msnp = microstacknetwork.packets
        self.service_funcs = {
            msnp.ACTION_NOP: lambda *args: None,
            msnp.ACTION_GET: self.action_get,
            msnp.ACTION_GET_BLOCK: self.action_get_block,
            msnp.ACTION_SET: self.action_set,
            msnp.ACTION_SET_BLOCK: self.action_set_block,
            msnp.ACTION_PROFILE: self.action_profile,
        }

    def handshake(self):
        for i in range(5):
            try:
                self.server_address, self.node_address = \
                    microstacknetwork.handshake.do_handshake(self.radio,
                                                             BEACON_ADDRESS,
                                                             NODE_INIT_ADDRESS)
            except microstacknetwork.handshake.HandshakeError as e:
                logging.error("NODE:handshake:{}".format(e))
                time.sleep(2)
            else:
                break
        else:
            logging.error("NODE:Failed all 5 attempts at handshake.")
            raise NodeFailedHandshakeError(
                "Failed all 5 attempts at handshake")

        logging.debug("NODE:completed handshake, my address is {}".format(
                      hex(self.node_address)))
        time.sleep(1)

    def run(self):
        # ask server for actions
        num_timeouts = 0
        interval = ACTION_REQ_INTERVAL
        while True:
            if num_timeouts >= 3:
                # listen for the beacon
                logging.error("NODE: Too many timeouts, listening for beacon.")
                self.radio.start_listening((BEACON_ADDRESS,))
                try:
                    pipe_num, data = self.radio.recv(timeout=10)
                except microstacknetwork.radios.nrf24l01.RXTimeoutError:
                    self.radio.stop_listening()
                    logging.error(
                        "NODE: Could not hear server beacon. Giving up!")
                    sys.exit(1)
                else:
                    self.radio.stop_listening()
                    num_timeouts = 0

            try:
                action_data = self.get_action_from_server()
            except (microstacknetwork.radios.RadioSendError,
                    microstacknetwork.radios.nrf24l01.RXTimeoutError) as e:
                logging.error("NODE:{}".format(e))
                interval = (ACTION_REQ_INTERVAL +
                            int(random.random() * 10) * ACTION_REQ_INTERVAL)
                num_timeouts += 1
            else:
                # success! so reset the interval and service the action
                interval = ACTION_REQ_INTERVAL
                self.service_action(action_data)
            finally:
                time.sleep(interval)

    def get_action_from_server(self):
        arpkt = microstacknetwork.packets.ActionRequestPacket()
        arpkt.src_address = self.node_address
        # input("Press enter to send action request packet.")
        logging.debug("NODE:Sending Action Request to {}."
                      .format(hex(self.server_address)))
        # self.radio.send(arpkt.to_bytes(), address=self.server_address)
        microstacknetwork.radios.radio_send(self.radio,
                                            3,
                                            arpkt.to_bytes(),
                                            self.server_address)

        self.radio.start_listening((self.node_address,))
        pipe_num, data = self.radio.recv(timeout=1)
        self.radio.stop_listening()

        return data

    def service_action(self, action_data):
        action = microstacknetwork.packets.ActionPacket()
        action.from_bytes(action_data)
        logging.debug("NODE:Received action:")
        logging.debug("NODE:{}".format(action))
        self.service_funcs[action.action_type](action_data)

    def action_profile(self, action_data):
        msnp = microstacknetwork.packets
        profile_action = msnp.ProfileActionPacket()
        profile_action.from_bytes(action_data)

        logging.debug("NODE:Building a profile.")
        channel_names = [c.name for c in self.channels]
        packets = msnp.profile_response(profile_action.action_id,
                                        channel_names)

        logging.debug("NODE:Sending profile. Packets:")
        try:
            for p in packets:
                logging.debug("NODE:{}".format(p))
                microstacknetwork.radios.radio_send(self.radio,
                                                    3,
                                                    p.to_bytes(),
                                                    self.server_address)
        except microstacknetwork.radios.RadioSendError:
            logging.error("NODE:Failed to send profile.")

    def action_get(self, action_data):
        msnp = microstacknetwork.packets
        get_action = msnp.GetActionPacket()
        get_action.from_bytes(action_data)

        logging.debug("NODE:Getting value.")
        value = self.channels[get_action.channel].get()

        get_response = msnp.GetResponsePacket(get_action.action_id, value)

        logging.debug("NODE:Sending response: {}".format(get_response))
        try:
            microstacknetwork.radios.radio_send(self.radio,
                                                3,
                                                get_response.to_bytes(),
                                                self.server_address)
        except microstacknetwork.radios.RadioSendError:
            logging.error("NODE:Could not send 'get' response.")

    def action_get_block(self, action_data):
        msnp = microstacknetwork.packets
        get_block_action = msnp.GetBlockActionPacket()
        get_block_action.from_bytes(action_data)

        logging.debug("NODE:Getting channel block (len: {}, start ch:{})."
                      .format(get_block_action.length,
                              get_block_action.start_channel))

        start = get_block_action.start_channel
        end = start + get_block_action.length
        try:
            values = self.channels[start].get_block(get_block_action.length)
        except NotImplementedError:
            values = [self.channels[i].get() for i in range(start, end)]

        packets = msnp.get_block_response(get_block_action.action_id, values)

        logging.debug("NODE:Sending get block response.")
        partial_send = False
        try:
            for p in packets:
                logging.debug("NODE:{}".format(p))
                microstacknetwork.radios.radio_send(self.radio,
                                                    3,
                                                    p.to_bytes(),
                                                    self.server_address)
                partial_send = True
        except microstacknetwork.radios.RadioSendError:
            logging.error("NODE:Could not send 'get block' response.")
            if partial_send:
                logging.error("NODE:I have sent some of the reponse though. "
                              "Has this been handled correctly on the server?")

    def action_set(self, action_data):
        msnp = microstacknetwork.packets
        set_action = msnp.SetActionPacket()
        set_action.from_bytes(action_data)

        logging.debug(
            "NODE:Setting channel {} to {}.".format(set_action.channel,
                                                    set_action.value))
        try:
            self.channels[set_action.channel].set(set_action.value)
        except KeyError:
            set_response = msnp.SetResponsePacket(set_action.action_id,
                                                  msnp.ERROR_NO_CHANNEL)
        else:
            set_response = msnp.SetResponsePacket(set_action.action_id,
                                                  msnp.SUCCESS)

        logging.debug("NODE:Sending response: {}".format(set_response))
        try:
            microstacknetwork.radios.radio_send(self.radio,
                                                3,
                                                set_response.to_bytes(),
                                                self.server_address)
        except microstacknetwork.radios.RadioSendError:
            logging.error("NODE:Could not send 'set' response.")

    def action_set_block(self, action_data):
        msnp = microstacknetwork.packets

        # first set_block_action
        set_block_action = msnp.SetBlockActionPacket()
        set_block_action.from_bytes(action_data)

        # Set Block Action packets
        sba_packets = [set_block_action]

        # while the last received packet is not the last packet in the sequence
        while not sba_packets[-1].is_last_packet():
            # receive more SetBlockActionPacket's
            self.radio.start_listening((self.node_address,))
            pipe_num, data = self.radio.recv(timeout=1)
            self.radio.stop_listening()
            set_block_action = msnp.SetBlockActionPacket()
            set_block_action.from_bytes(data)
            try:
                assert set_block_action.action_type == msnp.ACTION_SET_BLOCK
            except AssertionError:
                logging.error("NODE:set_block:Received a packet that was not "
                              "a set_block_action packet.")
            else:
                logging.node("NODE:Received a set_block_action packet.")
                sba_packets.append(set_block_action)

        start_channel = sba_packets[0].start_channel
        values = [p.values for p in sba_packets]  # [[1,2,3],[4,5]]
        values = sum(values, [])  # flatten the values: [1,2,3,4,5]

        logging.debug("NODE:Setting a block of {} channels from ch:{}."
                      .format(len(values), sba_packets[0].start_channel))
        for i, v in enumerate(values):
            self.channels[start_channel+i].set(v)

        set_block_response = \
            msnp.SetBlockResponsePacket(sba_packets[-1].action_id,
                                        msnp.SET_RESP_SUCCESS)

        logging.debug("NODE:Sending set block response.")
        try:
            microstacknetwork.radios.radio_send(self.radio,
                                                3,
                                                set_block_response.to_bytes(),
                                                self.server_address)
        except microstacknetwork.radios.RadioSendError:
            logging.error("NODE:Could not send 'set block' response.")


def degrees_and_minutes_to_degrees(degrees_and_minutes):
    """Converts dddmm.mmmm to ddd.dddd..."""
    degrees = int(degrees_and_minutes / 100)
    minutes = degrees_and_minutes % 100
    degrees += (minutes / 60)
    return degrees


if __name__ == '__main__':
    node = MicrostackNode()
    try:
        node.handshake()
        node.run()
    except KeyboardInterrupt:
        node.gps.stop()
